#!/usr/bin/env python3
"""
Support for alignment cigar-strings.

References
  [1] Sequence Alignment/Map Format Specification (the CIGAR field in
      section 1.4)
        https://samtools.github.io/hts-specs/SAMv1.pdf
"""


# cigar_to_counts--
#	Report the number of match, mismatch, and indel columns in the alignment
#	represented by a cigar string.
#
#   Note that the cigar string should have "X" and "=" operators instead of
#	simply "M".

def cigar_to_counts(cigar):
	cigarInfo = split_cigar(cigar)
	if (cigarInfo == None): return None

	numMatch = numMismatch = numIndel = 0

	for (rpt,op) in cigarInfo.operations:
		if (op == "="):
			numMatch += rpt
		elif (op == "X"):
			numMismatch += rpt
		elif (op in ["I","D"]):
			numIndel += rpt
		elif (op == "M"):
			raise ValueError("unsupported operation \"%d%s\" in cigar \"%s\""
			               + "\nonly cigar strings with operators in {=,X,I,D} are supported"
			               % (rpt,op,cigar))
		else:
			raise ValueError("unsupported operation \"%d%s\" in cigar \"%s\""
			               % (rpt,op,cigar))

	return (numMatch,numMismatch,numIndel)


# cigar_to_windowed_counts--
#	Report (yield) the number of match, mismatch, and indel columns in the
#	alignment represented by a cigar string, yielding the count for each
#	overlapping window in the reference sequence.
#
#   Note that the cigar string should have "X" and "=" operators instead of
#	simply "M".
#
#   Per reference [1],
#     "I" is a insertion to the reference; it consumes the reference but not the
#         query
#     "D" is a deletion from the reference; it consumes the query but not the
#         reference

def cigar_to_windowed_counts(cigar,windowSize,windowStep=1):
	cigarInfo = split_cigar(cigar)
	if (cigarInfo == None): return None

	# convert the cigar operations to a string

	opStr = []
	for (rpt,op) in cigarInfo.operations:
		if (op in ["=","X","I","D"]):
			opStr += [op*rpt]
		elif (op == "M"):
			raise ValueError("unsupported operation \"%d%s\" in cigar \"%s\""
			               + "\nonly cigar strings with operators in {=,X,I,D} are supported"
			               % (rpt,op,cigar))
		else:
			raise ValueError("unsupported operation \"%d%s\" in cigar \"%s\""
			               % (rpt,op,cigar))

	opStr = "".join(opStr)

	# assign in-string positions to each reference location

	refLocToIx = []
	refLoc = 0
	for (ix,ch) in enumerate(opStr):
		if (ch in ["=","X","I"]):
			refLocToIx += [ix]   # nota bene: refLocToIx[refLoc] = ix
			refLoc += 1
		else: # if (ch == "D"):
			pass
	refLocToIx += [len(opStr)]   # nota bene: refLocToIx[refLoc] = len(opStr)

	refLen = len(refLocToIx) - 1

	# extract window-sized substrings and count the operations within; note
	# that we strip any prefix or suffix deletion since an aligner would not
	# include those in an alignment of the interval

	refLoc = 0
	while (refLoc+windowSize <= refLen):
		ixLft = refLocToIx[refLoc]
		ixRgt = refLocToIx[refLoc+windowSize]
		windowStr = opStr[ixLft:ixRgt].strip("D")

		numMatch    = windowStr.count("=")
		numMismatch = windowStr.count("X")
		numIndel    = windowStr.count("I") + windowStr.count("D")
		yield (refLoc,numMatch,numMismatch,numIndel)

		refLoc += windowStep


# Earlier implementation of cigar_to_windowed_counts. This might be expected to
# be more efficient than "B", at least in terms of memory used.
#
# This seems to be correct but with one exception. If a window has a deletion
# prefix, this are unintentionally counted as part of the window's indels; I
# wasn't able to see an easy way to correct that. It seems like any time I
# yield after consuming an operation, I would need to look ahead to see if the
# next operation is a deletion.

#def cigar_to_windowed_counts_A(cigar,windowSize):
#	cigarInfo = split_cigar(cigar)
#	if (cigarInfo == None): return None
#	numOps = len(cigarInfo.operations)
#
#	numMatch = numMismatch = numIndel = 0
#	lftIx,lftRpt = (-1,0)
#	rgtIx,rgtRpt = (-1,0)
#	currentLen = 0   # length of current window, in the reference/target sequence
#	while (True):
#		# if we've exhausted the right-op fetch the next one, and validate it
#
#		if (rgtRpt == 0):
#			rgtIx += 1
#			if (rgtIx == numOps): break
#			(rgtRpt,rgtOp) = cigarInfo.operations[rgtIx]
#			if (rgtOp == "M"):
#				raise ValueError("unsupported operation \"%d%s\" in cigar \"%s\""
#							   + "\nonly cigar strings with operators in {=,X,I,D} are supported"
#							   % (rgtRpt,rgtOp,cigar))
#			elif (rgtOp not in ["=","X","I","D"]):
#				raise ValueError("unsupported operation \"%d%s\" in cigar \"%s\""
#							   % (rgtRpt,rgtOp,cigar))
#
#		# if we haven't yet accumulated a full window, absorb the right-op
#		# (some or all of it) into the first window; if as a result the window
#		# is full, yield the counts; note that once we fill the window this
#		# if clause will never trigger again
#
#		if (currentLen < windowSize):
#			shortfall = windowSize - currentLen
#			if (rgtOp == "="):
#				rptAbsorbed = min(rgtRpt,shortfall)
#				numMatch   += rptAbsorbed
#				rgtRpt     -= rptAbsorbed
#				currentLen += rptAbsorbed
#			elif (rgtOp == "X"):
#				rptAbsorbed = min(rgtRpt,shortfall)
#				numMismatch += rptAbsorbed
#				rgtRpt      -= rptAbsorbed
#				currentLen  += rptAbsorbed
#			elif (rgtOp == "I"):
#				# consumes reference only
#				rptAbsorbed = min(rgtRpt,shortfall)
#				numIndel   += rptAbsorbed
#				rgtRpt     -= rptAbsorbed
#				currentLen += rptAbsorbed
#			else: # if (rgtOp == "D"):
#				# consumes query only
#				numIndel += rgtRpt
#				rgtRpt = 0
#
#			if (currentLen == windowSize):
#				yield (..,numMatch,numMismatch,numIndel)
#			continue
#
#		# otherwise, the window is full, and we'll absorb an equal amount of
#		# left- and right-ops (equal relative to the reference sequence); note
#		# that we don't track currentLen here, because an invariant is that it
#		# will always be equal to the windowSize
#
#		# if we've exhausted the left-op fetch the next one; note that we don't
#		# need to validate it because we know we did so earlier when it was a
#		# right-op
#
#		if (lftRpt == 0):
#			lftIx += 1
#			(lftRpt,lftOp) = cigarInfo.operations[lftIx]
#
#		# determine how much to absorb from each op based on the op types; and
#		# for each base of the reference consumed, yield the counts
#
#		bothOp = lftOp+rgtOp
#		if (bothOp in ["==","=X","X=","XX"]):
#			rptAbsorbed = min(lftRpt,rgtRpt)
#			for _ in range(rptAbsorbed):
#				if (lftOp == "="): numMatch    -= 1
#				else:              numMismatch -= 1
#				if (rgtOp == "="): numMatch    += 1
#				else:              numMismatch += 1
#				yield (..,numMatch,numMismatch,numIndel)
#			lftRpt -= rptAbsorbed
#			rgtRpt -= rptAbsorbed
#		elif (bothOp == "II"):
#			rptAbsorbed = min(lftRpt,rgtRpt)
#			for _ in range(rptAbsorbed):
#				# (note that numIndel doesn't change)
#				yield (..,numMatch,numMismatch,numIndel)
#			lftRpt -= rptAbsorbed
#			rgtRpt -= rptAbsorbed
#		elif (bothOp == "DD"):
#			# note: no reference consumed, so no yield
#			numIndel -= lftRpt
#			numIndel += rgtRpt
#			lftRpt = rgtRpt = 0
#		elif (bothOp in ["=I","XI"]):
#			rptAbsorbed = min(lftRpt,rgtRpt)
#			for _ in range(rptAbsorbed):
#				if (lftOp == "="): numMatch    -= 1
#				else:              numMismatch -= 1
#				numIndel += 1
#				yield (..,numMatch,numMismatch,numIndel)
#			lftRpt -= rptAbsorbed
#			rgtRpt -= rptAbsorbed
#		elif (bothOp in ["I=","IX"]):
#			rptAbsorbed = min(lftRpt,rgtRpt)
#			for _ in range(rptAbsorbed):
#				numIndel -= 1
#				if (rgtOp == "="): numMatch    += 1
#				else:              numMismatch += 1
#				yield (..,numMatch,numMismatch,numIndel)
#			lftRpt -= rptAbsorbed
#			rgtRpt -= rptAbsorbed
#		elif (bothOp in ["=D","XD"]):
#			# note: no reference consumed, so no yield
#			numIndel += rgtRpt
#			rgtRpt =  0
#		elif (bothOp in ["D=","DX"]):
#			# note: no reference consumed, so no yield
#			numIndel -= lftRpt
#			lftRpt =  0
#		elif (bothOp == "ID"):
#			# note: no reference consumed, so no yield
#			numIndel += rgtRpt
#			rgtRpt =  0
#		elif (bothOp == "DI"):
#			# note: no reference consumed, so no yield
#			numIndel -= lftRpt
#			lftRpt =  0


# split_cigar--
#	Split a cigar string into a list of (count,operation)

class CigarInfo: pass

def split_cigar(cigar):
	if (cigar == "*"): return None

	# split the cigar into a list of (count,operation)

	operations = []
	rpt = []
	for ch in cigar:
		if (ch.isdigit()):
			rpt += [ch]
		else:
			#assert (rpt != []), "bad cigar: \"%s\"" % cigar
			if (rpt == []):
				operations += [(1,ch)]
			else:
				operations += [(int("".join(rpt)),ch)]
			rpt = []
	assert (rpt == []), "bad cigar: \"%s\"" % cigar

	# trim clipping operators from the ends

	startClip = endClip = 0
	if (operations != []):
		(rpt,op) = operations[0]
		if (op == "H"):
			startClip = rpt
			operations = operations[1:]

	if (operations != []):
		(rpt,op) = operations[-1]
		if (op == "H"):
			endClip = rpt
			operations = operations[:-1]

	splitCigar = CigarInfo()
	splitCigar.operations = operations
	splitCigar.startClip  = startClip
	splitCigar.endClip    = endClip
	return splitCigar


# cigar_to_string--
#	Convert a list of (count,operation) into a cigar string
#
# cigarInfo can be a CigarInfo object, or a list of (count,operation) pairs.
# Specifically, if it has an operations attribute, we assume it is a CigarInfo
# object; if not, we assume it is a list

def cigar_to_string(cigarInfo):
	startClip = endClip = 0
	if (hasattr(cigarInfo,"operations")):
		operations = cigarInfo.operations
		startClip  = cigarInfo.startClip
		endClip    = cigarInfo.endClip
	else:
		operations = cigarInfo

	operationsStr = "".join(["%s%s"%((str(rpt) if (rpt>1) else ""),op) for (rpt,op) in operations])
	if (startClip > 0): operationsStr = ("%dH"%startClip) + operationsStr
	if (endClip   > 0): operationsStr = operationsStr + ("%dH"%endClip)

	return operationsStr


# cigar_extent--
#	Determine the extent, the number of bases a cigar covers, in both sequences.
#
# (cigarInfo is as for cigar_to_string)

def cigar_extent(cigarInfo,targetStart=0,queryStart=0,reverseQuery=False):
	if (hasattr(cigarInfo,"operations")):
		operations = cigarInfo.operations
		#startClip  = cigarInfo.startClip  # these don't
		#endClip    = cigarInfo.endClip    # .. affect extent
	else:
		operations = cigarInfo

	qStep = -1 if (reverseQuery) else 1

	(tPos,qPos) = (targetStart,queryStart)
	for (rpt,op) in operations:
		if (op in ["M","X","="]):
			tPos += rpt
			qPos += qStep * rpt
		elif (op == "I"):
			qPos += qStep * rpt
		elif (op == "D"):
			tPos += rpt
		else:
			raise ValueError

	return (tPos,qPos)


# trace_cigar_path--
#	Return the list of (target,query) points in the path represented by a cigar
#	string.
#
# (cigarInfo is as for cigar_to_string)

def trace_cigar_path(cigarInfo,targetStart=0,queryStart=0,reverseQuery=False):
	if (hasattr(cigarInfo,"operations")):
		operations = cigarInfo.operations
		#startClip  = cigarInfo.startClip  # these don't
		#endClip    = cigarInfo.endClip    # .. affect the path
	else:
		operations = cigarInfo

	if (reverseQuery): 
		matchStep  = (1,-1)
		insertStep = (0,-1)
		deleteStep = (1,0)
	else:
		matchStep  = (1,1)
		insertStep = (0,1)
		deleteStep = (1,0)

	(tPos,qPos) = (targetStart,queryStart)
	path = [(tPos,qPos)]
	for (rpt,op) in operations:
		if (op in ["M","X","="]):
			(tStep,qStep) = matchStep
		elif (op == "I"):
			(tStep,qStep) = insertStep
		elif (op == "D"):
			(tStep,qStep) = deleteStep
		else:
			raise ValueError
		for _ in range(rpt):
			tPos += tStep
			qPos += qStep
			path += [(tPos,qPos)]

	return path


# cigar_before_target--
#	Return (as a list of (count,operation)) the prefix of a cigar string, up to
#	(but not including) a specified target position.
#
# (cigarInfo is as for cigar_to_string)

def cigar_before_target(cigarInfo,targetPos,targetStart=0):
	if (hasattr(cigarInfo,"operations")):
		operations = cigarInfo.operations
		#startClip  = cigarInfo.startClip  # these don't
		#endClip    = cigarInfo.endClip    # .. affect the path
	else:
		operations = cigarInfo

	# follow path until we reach the goal, collecting operations; the final
	# operation is shortened if necessary

	tPos = targetStart
	prefix = []
	for (rpt,op) in operations:
		if (tPos == targetPos):
			break
		assert (tPos < targetPos), "internal error in cigar_before_target, goal passed"
		if (op in ["M","X","=","D"]):
			if (tPos + rpt > targetPos):
				rpt = targetPos - tPos
			prefix += [(rpt,op)]
			tPos += rpt
		elif (op == "I"):
			prefix += [(rpt,op)]
		else:
			raise ValueError

	assert (tPos == targetPos), "internal error in cigar_before_target, goal not reached"

	return prefix


# cigar_after_target--
#	Return (as a list of (count,operation)) the suffix of a cigar string, after
#	(and including) a specified target position.
#
# (cigarInfo is as for cigar_to_string)

def cigar_after_target(cigarInfo,targetPos,targetStart=0):
	if (hasattr(cigarInfo,"operations")):
		operations = cigarInfo.operations
		#startClip  = cigarInfo.startClip  # these don't
		#endClip    = cigarInfo.endClip    # .. affect the path
	else:
		operations = cigarInfo

	# follow path until we reach the goal; the final operation is shortened
	# if necessary, with any leftover part recorded

	tPos = targetStart
	leftover = None
	for (ix,(rpt,op)) in enumerate(operations):
		if (tPos == targetPos):
			break
		assert (tPos < targetPos), "internal error in cigar_after_target, goal passed"
		if (op in ["M","X","=","D"]):
			if (tPos + rpt > targetPos):
				leftover = (tPos+rpt-targetPos,op)
				rpt = targetPos - tPos
			tPos += rpt
		elif (op == "I"):
			pass
		else:
			raise ValueError

	assert (tPos == targetPos), "internal error in cigar_after_target, goal not reached"

	# construct the suffix form the leftover (if any) and the remaining
	# operations

	suffix = [] if (leftover == None) else [leftover]
	suffix += operations[ix:]

	return suffix


# cigar_identity--
#	Compute the nucleotide identity (matches divided by matches+mismatches)
#	represented by a cigar string.
#
# (cigarInfo is as for cigar_to_string)

def cigar_identity(cigarInfo):
	if (hasattr(cigarInfo,"operations")):
		operations = cigarInfo.operations
		#startClip  = cigarInfo.startClip  # these don't
		#endClip    = cigarInfo.endClip    # .. affect the path
	else:
		operations = cigarInfo

	nMatch    = sum([rpt for (rpt,op) in operations if (op in ["M","="])])
	nMismatch = sum([rpt for (rpt,op) in operations if (op == "X")])
	return nMatch / (nMatch+nMismatch)


# cigar_blast_identity--
#	Compute the alignment identity including indels (matches divided by
#	matches+mismatches+indels) represented by a cigar string.
#
# (cigarInfo is as for cigar_to_string)

def cigar_blast_identity(cigarInfo):
	if (hasattr(cigarInfo,"operations")):
		operations = cigarInfo.operations
		#startClip  = cigarInfo.startClip  # these don't
		#endClip    = cigarInfo.endClip    # .. affect the path
	else:
		operations = cigarInfo

	nMatch     = sum([rpt for (rpt,op) in operations if (op in ["M","="])])
	nMutations = sum([rpt for (rpt,op) in operations if (op in ["X","I","D"])])
	return nMatch / (nMatch+nMutations)


# cigar_continuity--
#	Compute the continuity (the fraction of alignment columns that do not
#	contain gaps) represented by a cigar string.
#
# (cigarInfo is as for cigar_to_string)

def cigar_continuity(cigarInfo):
	if (hasattr(cigarInfo,"operations")):
		operations = cigarInfo.operations
		#startClip  = cigarInfo.startClip  # these don't
		#endClip    = cigarInfo.endClip    # .. affect the path
	else:
		operations = cigarInfo

	nMatchOrSub = sum([rpt for (rpt,op) in operations if (op in ["M","=","X"])])
	nColumns    = sum([rpt for (rpt,op) in operations])
	return nMatchOrSub / nColumns


# cigar_nmatch--
#	Count the number of matched bases represented by a cigar string.
#
# (cigarInfo is as for cigar_to_string)

def cigar_nmatch(cigarInfo):
	if (hasattr(cigarInfo,"operations")):
		operations = cigarInfo.operations
		#startClip  = cigarInfo.startClip  # these don't
		#endClip    = cigarInfo.endClip    # .. affect the path
	else:
		operations = cigarInfo

	nMatch = sum([rpt for (rpt,op) in operations if (op in ["M","="])])
	return nMatch




"""
Note: the initial implementation of this predates pyfaidx. I maintain this to
      support backward compatibility for old scripts. But new scripts will do
      much better by using pyfaidx instead of this.
"""

from collections import MutableMapping
from os.path     import exists
from sys         import stderr

try:
	from hashlib import md5 as md5_new
except ImportError:
	try:
		from md5 import new as md5_new
	except ImportError:
		md5_new = None


class FastaFile(MutableMapping):

	def __init__(self,f,unmask=False,requireFai=False,ignoreFai=False,debug=None):
		"""
		f can be either a file object or a filename
		"""
		self.debug = debug if (debug != None) else []

		if (requireFai) and (ignoreFai):
			raise ValueError(".fai is required but is also to be ignored(!)")
		if (requireFai) and (type(f) != str):
			raise ValueError(".fai is required but fai filename cannot be constructed")

		fai = None
		if (type(f) == str) and ((requireFai) or (not ignoreFai)):
			if (f.endswith(".bgz")) or (f.endswith(".gz")) or (f.endswith(".bz2")) or (f.endswith(".zip")):
				raise ValueError("this .fai implementation doesn't support compressed fasta files (\"%d\")" % f)
			faiFilename = f+".fai"
			faiExists = exists(faiFilename)
			if (faiExists):
				fai = open(faiFilename,"r")
			elif (requireFai):
				raise ValueError(".fai is required but \"%d\" doesn't exist" % faiFilename)

		if (fai != None):
			f = open(f,"rb")
		elif (type(f) == str):
			f = open(f,"r")

		self.file   = f
		self.unmask = unmask

		self.seqCount = 0
		self.index = {}
		self.alias = {}

		if (fai == None):
			# read the sequences and create a name-to-sequence hash; we also
			# create an number-to-name hash; and if there is only one sequence,
			# we will map None to that name

			for (name,seq) in self.read_all_sequences(f):
				if (name in self.index):
					raise Exception("sequence name \"%s\" appears more than once in fasta file)" \
								  % (name,nameIx,nameOffset))
				if (unmask): seq = seq.upper()
				self.index[name] = seq
				self.alias[self.seqCount] = name
				self.seqCount += 1
				if ("signatures" in self.debug):
					seqHash = md5_new()
					seqHash.update(seq.encode("utf-8"))
					print("signature(\"%s\") = %s" % (name,seqHash.hexdigest().upper()),file=stderr)
			if (self.seqCount == 1): self.alias[None] = self.alias[0]
		else:
			# read the sequences and create a name-to-index-info hash (as well
			# as the number-to-name hash); entries in this hash will be
			# replaced by the sequence on an as-needed basis

			for (name,length,offset,lineBases,lineBytes) in self.read_fai(fai):
				if (name in self.index):
					raise Exception("sequence name \"%s\" appears more than once in fai file)" \
								  % (name,nameIx,nameOffset))
				self.index[name] = (length,offset,lineBases,lineBytes)
				self.alias[self.seqCount] = name
				self.seqCount += 1
			if (self.seqCount == 1): self.alias[None] = self.alias[0]

	def __getitem__(self,name):
		try:
			# convert to alias if necessary, and lookup sequence -or- fai info
			if (name not in self.index): name = self.alias[name]
			seq = self.index[name]
		except KeyError:
			return None
		if (type(seq) == tuple):
			# if we have fai info this sequence was not previously loaded; do
			# so now
			(length,offset,lineBases,lineBytes) = seq
			seq = self.read_sequence(self.file,length,offset,lineBases,lineBytes)
			if (self.unmask): seq = seq.upper()
			self.index[name] = seq
			if ("signatures" in self.debug):
				seqHash = md5_new()
				seqHash.update(seq.encode("utf-8"))
				print("signature(\"%s\") = %s" % (name,seqHash.hexdigest().upper()),file=stderr)
		return seq

	def __setitem__(self,name,val):
		raise Exception("FastaFile doesn't support __setitem__)")

	def __delitem__(self,name):
		raise Exception("FastaFile doesn't support __delitem__)")

	def __len__(self):
		return len(self.index)

	def __iter__(self):
		for name in self.index:
			yield name

	def keys(self):
		return self.index.keys()

	def read_all_sequences(self,f):
		seqName = None
		lineNum = 0
		for line in f:
			lineNum += 1
			line = line.strip()

			if (line.startswith(">")):
				if (seqName != None):
					seq = "".join(seq)
					yield (seqName,seq)
				fields = line[1:].split()
				if (fields == []):
					assert (False), \
					       "sequence has no name (at line %d)" % lineNum
				seqName = fields[0]
				seq = []
			elif (seqName == None):
				assert (False), "first sequence has no header"
			else:
				seq += [line]

		if (seqName != None):
			seq = "".join(seq)
			yield (seqName,seq)

	def read_sequence(self,f,length,offset,lineBases,lineBytes):
		seq = []
		lineOffset = offset
		basesLeft = length
		while (basesLeft > 0):
			f.seek(lineOffset)
			basesOnLine = lineBases if (basesLeft >= lineBases) else basesLeft
			seq += [f.read(basesOnLine).decode()]
			lineOffset += lineBytes
			basesLeft  -= basesOnLine
		return "".join(seq)

	def read_fai(self,f):
		# see http://www.htslib.org/doc/faidx.html
		lineNumber = 0
		for line in f:
			lineNumber += 1
			line = line.strip()

			fields = line.split()
			try:
				if (len(fields) != 5): raise ValueError
				name      = fields[0]
				length    = int(fields[1])
				offset    = int(fields[2])
				lineBases = int(fields[3])
				lineBytes = int(fields[4])
				if (length < 0): raise ValueError
				if (offset < 0): raise ValueError
				if (lineBases < 0): raise ValueError
				if (lineBytes < lineBases): raise ValueError
			except ValueError:
				raise ValueError("improper fai line (%d)\n%s" % (lineNumber,line))

			yield (name,length,offset,lineBases,lineBytes)




#!/usr/bin/env python3
"""
Read pairwise alignments from lastz's general format. Typical input is below
but other columns may be included. It is up to the calling program what
columns are required.

    #name1    zstart1  end1     name2            strand2 zstart2+ end2+  id
    mule.chr1 9833952  9834016  donkey.contig12  -       142287   142351 97.7
    mule.chr1 8336423  8336491  donkey.contig12  +       30291    30359  95.9
    mule.chr1 25360782 25360844 donkey.contig12  -       160824   160886 98.6
    mule.chr1 11648921 11648978 donkey.contig163 -       51135    51192  95.9
    mule.chr1 11283997 11284065 donkey.contig163 -       57665    57733  99.0
    mule.chr1 18960023 18960091 donkey.contig163 +       56549    56617  96.9
     ...
"""

class Alignment: pass

class AlignmentTable(object):

	def __init__(self):
		"""
		We don't expet direct calls to this constructor. Instead, we expect
		objects to be created by calls to from_file().
		"""
		pass

	def __iter__(self):
		return self

	def __next__(self):
		if (self.iteratorStopped):
			if (self.filename != None): self.f.close()
			raise StopIteration
		elif (self.iteratorLookahead != None):
			a = self.iteratorLookahead
			self.iteratorLookahead = None
		else:
			try:
				a = self.iterator.__next__() # may raise StopIteration
			except StopIteration:
				if (self.filename != None): self.f.close()
				raise StopIteration
		return a

	@classmethod
	def from_file(cls,f,preFill=False,columnNames=None,
	              requiredColumns=None,nonRequiredColumns=None,columnAliases=None):
		"""
		Create an AlignmentTable from a file or stream.

		If preFill is True, we read the entire file into t.alignments.

		Otherwise (if preFill is False) we set t up as an iterator. The caller
		can then use the object like this:
		  t = AlignmentTable.from_file(...)
		  if (t.header != None):
		    do something with t.header
		  for a in t:
		    do something with alignment a

		f can either be a file object (already opened for read) or the name of
		a file (which we will open, read, and close).

		columnName, requiredColumns, nonRequiredColumns, and columnAliases are as
		defined in read_file().
		"""

		filename = None
		if (type(f) == str):
			filename = f
			if (filename.endswith(".gz")) or (filename.endswith(".gzip")):
				from gzip import open as gzip_open
				f = gzip_open(filename,"rt")
			else:
				f = open(filename,"rt")

		t = AlignmentTable()
		t.f = f
		t.filename = filename
		t.header = None
		t.columnNames = columnNames

		# pre-fill the table

		if (preFill):
			t.iterator   = None
			t.alignments = []
			for a in AlignmentTable.read_file(f,columnNames=columnNames,
			                                  requiredColumns=requiredColumns,
			                                  nonRequiredColumns=nonRequiredColumns,
			                                  columnAliases=columnAliases):
				if (type(a) == str):
					t.header = a
				elif (type(a) == dict):
					if (t.columnNames == None):
						t.columnNames = a
				else:
					t.alignments += [a]
			if (filename != None):
				f.close()

		# don't pre-fill, set up an iterator

		else:
			t.alignments = None
			t.iterator   = AlignmentTable.read_file(f,columnNames=columnNames,
			                                        requiredColumns=requiredColumns,
			                                        nonRequiredColumns=nonRequiredColumns,
			                                        columnAliases=columnAliases)

			# fetch the first item, which may be the header

			try:
				a = t.iterator.__next__()
				t.iteratorStopped = False
			except StopIteration:
				t.iteratorStopped = True

			if (type(a) == str):
				t.header = a
				t.iteratorLookahead = None
			else:
				t.iteratorLookahead = a

			# fetch the second item, which may be the column names map

			if (not t.iteratorStopped):
				if (t.iteratorLookahead != None):
					a = t.iteratorLookahead
					t.iteratorLookahead = None
				else:
					try:
						a = t.iterator.__next__()
						t.iteratorStopped = False
					except StopIteration:
						t.iteratorStopped = True

				if (type(a) == dict):
					if (t.columnNames == None): t.columnNames = a
					t.iteratorLookahead = None
				else:
					t.iteratorLookahead = a

		return t


	@classmethod
	def read_file(cls,f,columnNames=None,requiredColumns=None,nonRequiredColumns=None,columnAliases=None):
		"""
		f is a file object (already opened for read).

		If columnNames is None, we require the first line of the file to be
		column headers, and extract the names from those. If columnNames is
		*not* None, the first line may or may not be column headers. If it is
		column headers, they must match what has been given in columnNames.
		columnNames is either a list of names, or a dict mapping a name to a
		column number (0 is the first column).

		requiredColumns lists the names of columns that are required to be
		present. The first six required columns *must* correspond to the fields
		name1, zstart1, end1, name2, zstart2+, and end2+. They don't have to
		have those specific names, but the must be conceptually equivalent to
		them.

		nonRequiredColumns lists the names of additional columns that, if
		present, will be carried as attributes in the alignment objects.

		columnAliases is a dict than maps an alias to an equivalent column
		name. For example, the input file might have a column "s" instead of
		"strand2".

		Note that we yield three types of objects:
		  - a string, which is the header line (if there is one)
		  - a dict, which maps names to column numbers (equivalent to
		    columnNames)
		  - a series of Alignments
		"""
		if (type(columnNames) == list):
			columnNames = {name:ix for (ix,name) in enumerate(columnNames)}

		if (requiredColumns == None):
			requiredColumns = ("name1","zstart1","end1",
			                   "name2","zstart2+","end2+",
			                   "strand")
		else:
			if (len(requiredColumns) < 6):
				raise ValueError("at least six columns must be specified in requiredColumns")
			requiredColumns = tuple(requiredColumns)

		if (columnAliases == None):
			columnAliases = {"strand2" : "strand",
			                 "s"       : "strand",
			                 "s2"      : "strand",
			                 "text1"   : "align1",
			                 "text2"   : "align2",
			                 "id%"     : "id",
			                 "con%"    : "con",
			                 "cov1%"   : "cov1",
			                 "cov2%"   : "cov2"}

		headerNames  = None
		name1Column  = None
		start1Column = None
		end1Column   = None
		name2Column  = None
		start2Column = None
		end2Column   = None

		numColumnsNeeded = None

		header = None
		lineNumber = 0
		for line in f:
			lineNumber += 1
			line = line.strip()

			# process the header line

			if (line.startswith("#")):
				if (header != None): continue  # we ignore all but the first header line
				header = line[1:]
				fields = line.split()
				fields[0] = fields[0][1:]
				headerNames = AlignmentTable.column_names(fields,requiredColumns,nonRequiredColumns=nonRequiredColumns,columnAliases=columnAliases)
				if (columnNames != None):
					differences = AlignmentTable.column_name_differences(headerNames,columnNames)
					if (differences != None):
						differences = "\n".join(["%s %d in file but spcified as %d"%(name,1+hIx,1+cIx) for (name,hIx,cIx) in differences])
						raise ValueError("column names provided within the file don't match those specified\n%s" % differences)
				else:
					columnNames = headerNames
					yield line
					yield columnNames
				continue

			# if we needed a header line and the input doesn't have one, it's
			# an error

			if (columnNames == None):
				raise ValueError("input column names weren't specified, and are not provided within the file")

			# on the arrival of the first non-header, derive some column
			# mapping info

			if (numColumnsNeeded == None):
				numColumnsNeeded = 1 + max([columnNames[name] for name in columnNames])

				(name1Column,start1Column,end1Column,
	             name2Column,start2Column,end2Column) = requiredColumns[:6]

				name1Column  = columnNames[name1Column ]
				start1Column = columnNames[start1Column]
				end1Column   = columnNames[end1Column  ]
				name2Column  = columnNames[name2Column ]
				start2Column = columnNames[start2Column]
				end2Column   = columnNames[end2Column  ]

			# split the line into fields

			fields = line.split()
			if (len(fields) < numColumnsNeeded):
				raise ValueError("not enough columns at line %d (%d, expected at least %d)" \
				               % (lineNumber,len(fields),numColumnsNeeded))

			# create the Alignment object and populate the common attributes

			a = Alignment()
			a.lineNumber = lineNumber
			a.line       = line
			a.name1      = fields[name1Column ]
			a.start1     = fields[start1Column]
			a.end1       = fields[end1Column  ]
			a.name2      = fields[name2Column ]
			a.start2     = fields[start2Column]
			a.end2       = fields[end2Column  ]

			# populate the remaining attributes

			for name in requiredColumns[6:]:
				if (name in columnAliases): name = columnAliases[name]
				val = fields[columnNames[name]]
				# $$$ perhaps recognize some fields and parse them, such as:
				# if (name == "score"): val = int_or_float(val)
				setattr(a,name,val)

			if (nonRequiredColumns != None):
				for name in nonRequiredColumns:
					if (name in columnAliases): name = columnAliases[name]
					if (name not in columnNames): continue
					val = fields[columnNames[name]]
					# $$$ perhaps recognize some fields and parse them, such as:
					# if (name == "score"): val = int_or_float(val)
					setattr(a,name,val)

			# validate the positional attributes
			# $$$ add a way to flip start2 and end2 for negative strand

			try:
				a.start1 = int(a.start1)
				a.end1   = int(a.end1)
				if (a.start1 >= a.end1): raise ValueError
			except ValueError:
				raise ValueError("bad alignment (at line %d), first species start/end\n%s" \
				               % (lineNumber,line))

			try:
				a.start2 = int(a.start2)
				a.end2   = int(a.end2)
				if (a.start2 >= a.end2): raise ValueError
			except ValueError:
				raise ValueError("bad alignment (at line %d), second species start/end\n%s" \
				               % (lineNumber,line))

			if ("strand" in requiredColumns):
				try:
					if (a.strand not in ["+","-"]): raise ValueError
				except ValueError:
					raise ValueError("bad alignment (at line %d), bad strand \"%s\"\n%s" \
					               % (lineNumber,a.strand,line))

			yield a

	@staticmethod
	def column_names(names,requiredColumns,nonRequiredColumns=None,columnAliases=None):
		if (columnAliases == None): columnAliases = {}
		namesOfInterest = set(requiredColumns)
		if (nonRequiredColumns != None):
			namesOfInterest.update(nonRequiredColumns)
		columnNames = {}
		for (ix,name) in enumerate(names):
			actualName = name
			if (name in columnAliases): name = columnAliases[name]
			if (name not in namesOfInterest): continue
			if (name in columnNames):
				raise ValueError(("\"%s\" (or alias) appears more than once in --format" % actualName))
			columnNames[name] = ix
		for name in requiredColumns:
			if (name not in columnNames):
				raise ValueError(("required name \"%s\" is absent" % name))
		return columnNames

	@staticmethod
	def column_name_differences(columnNames1,columnNames2):
		differences = []
		for name in columnNames2:
			if (name not in columnNames1):
				differences += [(name,-1,columnNames2[name])]
			if (columnNames1[name] != columnNames2[name]):
				differences += [(name,columnNames1[name],columnNames2[name])]
		for name in columnNames1:
			if (name not in columnNames2):
				differences += [(name,columnNames1[name],-1)]
		return None if (differences == []) else differences



# class supporting reads from UCSC's .2bit format for DNA sequence collections.
#
# The .2bit file format spec can be found at
#   http://genome.ucsc.edu/FAQ/FAQformat#format7
#
# History: this was originally written circa 2005 by Bob Harris, and some
# version of it was incorporated in James Taylor's bx python package.

import sys

from struct      import *
from collections import MutableMapping
from sys         import stderr


class TwoBitFile(MutableMapping):

	TWOBIT_MAGIC_NUMBER      = 0x1A412743
	TWOBIT_MAGIC_NUMBER_SWAP = 0x4327411A
	TWOBIT_VERSION           = 0

	byteToDna = [None] * 256
	for ix in range(256):
		nuc1 = (ix >> 6)
		nuc2 = (ix >> 4) & 3
		nuc3 = (ix >> 2) & 3
		nuc4 =  ix       & 3
		byteToDna[ix] = ["TCAG"[nuc1],"TCAG"[nuc2],"TCAG"[nuc3],"TCAG"[nuc4]]

	def __init__(self,f,unmask=False,debug=None):
		self.debug = debug if (debug != None) else []

		# read magic and determine byte order

		if (type(f) == str):
			f = open(f,"rb")

		magic = unpack(">L",f.read(4))[0]
		if (magic == TwoBitFile.TWOBIT_MAGIC_NUMBER):
			self.byteOrder = ">"
		elif (magic == TwoBitFile.TWOBIT_MAGIC_NUMBER_SWAP):
			self.byteOrder = "<"
		else:
			raise Exception("Not a 2bit file")

		self.file   = f
		self.unmask = unmask

		self.version = self.read("L")
		if (self.version != TwoBitFile.TWOBIT_VERSION):
			raise Exception("unsupported 2bit version %d" % self.version)

		# read the sequence names and create a hash

		self.seqCount = self.read("L")
		reserved      = self.read("L")
		if (reserved != 0):
			raise Exception("2bit file has %d at offset 0x0000000C (supposed to be zero)" \
			              % reserved)
		readOffset = 0x0000000C + 4

		self.index = {}
		for nameIx in range(self.seqCount):
			nameOffset = readOffset
			name = self.read_p_string()
			headerOffset = self.read("L")
			readOffset += 1 + len(name) + 4
			if (name in self.index):
				raise Exception("sequence name \"%s\" appears more than once in 2bit file index (fileIndex[%d] at file offset 0x%08X)" \
				              % (name,nameIx,nameOffset))
			self.index[name] = headerOffset  # this will be replaced by seq object if/when fetched
		self.preloadedCount = 0

		self.len = {}
		for name in self.index:
			self.len[name] = self.sequence_length(name)

	def __getitem__(self,name):
		try:
			seq = self.index[name]
			if (type(seq) == int):
				self.index[name] = seq = TwoBitSequence(self,seq)
			if (seq.ready): return seq
			self.preload_sequence(name)
			return seq
		except KeyError:
			return None

	def __setitem__(self,name,val):
		raise Exception("TwoBitFile doesn't support __setitem__)")

	def __delitem__(self,name):
		raise Exception("TwoBitFile doesn't support __delitem__)")

	def __len__(self):
		return len(self.index)

	def __iter__(self):
		for name in self.index:
			yield name

	def keys(self):
		return self.index.keys()

	def sequence_length(self,name):
		seq = self.index[name]
		if (type(seq) != int):
			# the sequence has been preloaded; ask it for its length
			return len(seq)
		else:
			# the sequence has not been preloaded; read its length from file
			headerOffset = seq
			self.file.seek(headerOffset)
			return self.read("L")

	def preload_sequence(self,name):
		seq = self.index[name]
		if (type(seq) == int):
			self.index[name] = seq = TwoBitSequence(self,seq)

		self.file.seek(seq.headerOffset)
		seq.size = self.read("L")
		if ("blocks" in self.debug):
			print ("reading blocks for \"%s\" (%d/%d)" \
			     % (name,self.preloadedCount+1,self.seqCount),
			       file=stderr)
		nBlocks    = self.read_blocks_info()
		maskBlocks = self.read_blocks_info()
		reserved   = self.read("L")
		if (reserved != 0):
			raise Exception("2bit file has %d at offset 0x%08X (supposed to be zero)" \
			              % (reserved,self.file.tell()-4))
		seq.sequenceOffset = self.file.tell()

		seq.nBlocks = nBlocks
		if (self.unmask): seq.maskBlocks = []
		else:             seq.maskBlocks = maskBlocks

		seq.ready = True
		self.preloadedCount += 1

	def read_blocks_info(self):
		blockCount = self.read("L")
		if (blockCount == 0): return []
		starts = self.read(str(blockCount)+"L",untuple=False)
		sizes  = self.read(str(blockCount)+"L",untuple=False)
		blocks = [(start,start+size) for (start,size) in zip(starts,sizes)]
		blocks.sort()
		return blocks

	def read(self,pattern,untuple=True):
		pattern = self.byteOrder + pattern
		val = unpack(pattern,self.file.read(calcsize(pattern)))
		if (untuple) and (len(val) == 1): return val[0]
		else:                             return val

	def read_p_string(self):
		length = self.read("B")
		return self.file.read(length).decode("utf-8")


class TwoBitSequence(object):
	def __init__(self,tbf,headerOffset):
		self.tbf            = tbf
		self.headerOffset   = headerOffset
		self.size           = None
		self.nBlocks        = None  # sorted list of (start,end)
		self.maskBlocks     = None  # sorted list of (start,end)
		self.sequenceOffset = None
		self.ready          = False

	def __getitem__(self,slice):
		if ("getitem" in self.tbf.debug):
			print ("slice = %s" % slice,file=stderr)
		(start,stop,stride) = slice.indices(self.size)
		assert (stride == 1), "Striding in slices not supported"
		if (stop <= start): return ""
		return self.read_slice(start,stop)

	def __len__(self):
		return self.size

	def get(self,start,end):
		if (start < 0):         start = 0
		if (end   > self.size): end  = self.size
		if (stop <= start): return ""
		return self.read_slice(start,stop)

	def read_slice(self,start,end):
		# $$$ we could improve this by checking the nBlocks first to see if we
		# $$$ .. will be N-masking the whole slice;  save the relevant part of
		# $$$ .. that list (if any), read, then apply

		if ("slice" in self.tbf.debug):
			print("start,end = %s,%s" % (start,end),file=stderr)

		nucsToFetch = end - start

		packedStart = start >> 2
		packedEnd   = (end+3) >> 2
		packedSize  = packedEnd - packedStart

		self.tbf.file.seek(self.sequenceOffset + packedStart)
		packedBytes = self.tbf.file.read(packedSize)

		if (packedSize == 1):
			# dna is entirely within one packed byte
			sOff = start - (packedStart << 2)
			eOff = sOff  + nucsToFetch
			packed   = packedBytes[0]
			unpacked = TwoBitFile.byteToDna[packed]
			dna = unpacked[sOff:eOff]
		else:
			# dna spans many packed bytes
			dna = []
			packedIx = 0

			skip = start & 3
			if (skip > 0):
				keep = 4 - skip
				packed   = packedBytes[packedIx]
				unpacked = TwoBitFile.byteToDna[packed]
				packedIx += 1
				dna += unpacked[skip:]
				nucsToFetch -= keep

			while (nucsToFetch >= 4):
				packed = packedBytes[packedIx]
				dna    += TwoBitFile.byteToDna[packed]
				packedIx += 1
				nucsToFetch -= 4

			if (nucsToFetch > 0):
				packed   = packedBytes[packedIx]
				unpacked = TwoBitFile.byteToDna[packed]
				dna += unpacked[:nucsToFetch]

		# apply N- and mask-intervals

		for (s,e) in self.nBlocks:
			if (e <= start): continue
			if (s >= end):   break
			if (s < start):  s = start
			if (e > end):    e = end
			s -= start
			e -= start
			dna[s:e] = ["N"] * (e-s)

		for (s,e) in self.maskBlocks:
			if (e <= start): continue
			if (s >= end):   break
			if (s < start):  s = start
			if (e > end):    e = end
			s -= start
			e -= start
			dna[s:e] = [ch for ch in "".join(dna[s:e]).lower()]

		return "".join(dna)




#!/usr/bin/env python3
"""
Things built upon the AlignmentTable objects.
"""

# construct_alignment_text--
#	Construct the alignment text, tracing a cigar string's path through the
#	sequences.

def construct_alignment_text(a,cigar,targetLookup,queryLookup):
	tSubstring = targetLookup[a.name1][a.start1:a.end1]
	qSubstring = queryLookup [a.name2][a.start2:a.end2]
	if (a.strand == "-"): qSubstring = reverse_complement(qSubstring)

	text1 = []
	text2 = []

	tIx = qIx = 0
	for (rpt,op) in cigar.operations:
		if (op in ["M","X","="]):
			text1 += tSubstring[tIx:tIx+rpt]
			text2 += qSubstring[qIx:qIx+rpt]
			tIx += rpt
			qIx += rpt
		elif (op == "I"):
			text1 += ["-"] * rpt
			text2 += qSubstring[qIx:qIx+rpt]
			qIx += rpt
		elif (op == "D"):
			text1 += tSubstring[tIx:tIx+rpt]
			text2 += ["-"] * rpt
			tIx += rpt
		else:
			assert (False), "(at line %s) unsupported \"%d%s\" in cigar \"%s\"" \
			              % (("{:,}".format(a.lineNumber)),rpt,op,a.cigarx)

	if (tIx != a.end1 - a.start1):
		assert (False), ("(at line %s) internal error(?);" 
		               + " target %s interval (%d,%d) doesn't match cigar ends (%d,%d)") \
		              % (("{:,}".format(a.lineNumber)),a.name1,a.start1,a.end1,a.start1,a.start1+tIx)

	if (qIx != a.end2 - a.start2):
		assert (False), ("(at line %s) internal error(?);" 
		               + " query %s interval (%d,%d) doesn't match cigar ends (%d,%d)") \
		              % (("{:,}".format(a.lineNumber)),a.name2,a.start2,a.end2,a.start2,a.start2+qIx)

	return ("".join(text1),"".join(text2))


# alignment_score--
#	Compute the score, using lastz default scoring, of an alignment. The
#	alignment is provided as the two aligning strings text1 and text2.
#
# Nota bene: per https://lastz.github.io/lastz/#options_scoring, the first base
#            in a gap incurs the sum of gap open and extend penalties.

# $$$ ideally we'd like to be able to use other scoring schemes, read scores
#     .. from files, etc.

def alignment_score(text1,text2):
	gapOpenPenalty    = 400     # (see note
	gapExtendPenalty  = 30      #  .. above)
	xPenalty          = 1000
	nPenalty          = 100
	substitutionScore = {"AA":   91, "AC":-114, "AG": -31, "AT":-123,
	                     "CA": -114, "CC": 100, "CG":-125, "CT": -31,
	                     "GA":  -31, "GC":-125, "GG": 100, "GT":-114,
	                     "TA": -123, "TC": -31, "TG":-114, "TT":  91}

	# $$$ there's bound to be a faster way to do this

	score = 0
	gapLen1 = gapLen2 = 0
	for (ch1,ch2) in zip(text1.upper(),text2.upper()):
		if (ch1 != "-") and (ch2 != "-"):
			if (gapLen1 > 0):
				score -= gapOpenPenalty + gapLen1*gapExtendPenalty
				gapLen1 = 0
			elif (gapLen2 > 0):
				score -= gapOpenPenalty + gapLen2*gapExtendPenalty
				gapLen2 = 0
			key = ch1+ch2
			if (key in substitutionScore):
				score += substitutionScore[key]
			elif (ch1 == "X") or (ch2 == "X"):
				score -= xPenalty
			else:
				score -= nPenalty
		elif (ch1 == "-"):
			if (gapLen2 > 0):
				score -= gapOpenPenalty + gapLen2*gapExtendPenalty
				gapLen2 = 0
			gapLen1 += 1
		elif (ch2 == "-"):
			if (gapLen1 > 0):
				score -= gapOpenPenalty + gapLen1*gapExtendPenalty
				gapLen1 = 0
			gapLen2 += 1
		else:
			pass  # both nts are a gap -- just ignore it

	if (gapLen1 > 0):
		score -= gapOpenPenalty + gapLen1*gapExtendPenalty
	elif (gapLen2 > 0):
		score -= gapOpenPenalty + gapLen2*gapExtendPenalty

	return score


# reverse_complement--

complementMap = str.maketrans("ACGTSWRYMKBDHVNacgtswrymkbdhvn",
                              "TGCASWYRKMVHDBNtgcaswyrkmvhdbn")

def reverse_complement(nukes):
	return nukes[::-1].translate(complementMap)



#!/usr/bin/env python3
"""
Convert pairwise alignments from lastz's general format to MAF.

References
  [1] MAF format (specification)
        http://genome.ucsc.edu/FAQ/FAQformat#format5
"""

from sys                  import argv,stdin,stderr,exit
from math                 import ceil
from alignment_table      import AlignmentTable
from alignment_table_aids import construct_alignment_text,alignment_score
from cigar                import split_cigar

try:                from two_bit_file import TwoBitFile
except ImportError: TwoBitFile = None

try:                from fasta_file import FastaFile
except ImportError: FastaFile = None


programName    = "tabular_to_maf"
programVersion = "0.1.0"


def usage(s=None):
	message = """
usage: cat <alignment_file> | %s [options]
  --sequence[s]=<file>    (required) specify the reference sequence(s); <file>
						  can be .2bit, .fasta, or .fasta.gz; for .fasta, if an
						  associated .fai file exists, it is utilized
  --alias:<alias>=<name>  the input can use <alias> as an alias for column
                          name <name>
  --head=<number>         limit the number of alignment records
  --progress=<number>     periodically report how many alignment records we've
                          processed
  --version               report this program's version number

Read pairwise alignments from lastz's general format and, using the cigar
strings, create a dotplot table suitable for plotting in R.

Typical input is shown below but other columns may be included, and columns can
appear in any order. Note that cigar or cigarx (either one) is required. If
score is included, it is used; otherwise the score is re-calculated (using
lastz defaults). If text1 and/or text2 are both included, they are used;
otherwise they are reconstructed using the cigar string. 

  #name1    zstart1  end1     name2            strand2 zstart2+  end2+     cigarx
  mule.chr1 1238615  1650901  donkey.contig12  +       1183966   1594899   ...
  mule.chr1 1952507  2309884  donkey.contig12  +       1889643   2247263   ...
  mule.chr1 3530276  3888853  donkey.contig12  +       88799552  89157262  ...
  mule.chr1 62122793 62459747 donkey.contig163 +       154012987 154349815 ...
  mule.chr1 4856000  5077740  donkey.contig163 +       90480442  90701655  ...
  mule.chr1 3926177  4143832  donkey.contig163 +       89190167  89407555  ...
   ...

Output is in MAF format, as described at
  http://genome.ucsc.edu/FAQ/FAQformat#format5""" \
% programName

	if (s == None): exit (message)
	else:           exit ("%s\n%s" % (s,message))


def main():
	global debug

	fieldJoiner = "\t"

	# parse the command line

	referenceFilenames = []
	aliasToName        = {}
	headLimit          = None
	reportProgress     = None
	debug              = []
	refDebug           = []

	for arg in argv[1:]:
		if ("=" in arg):
			argVal = arg.split("=",1)[1]

		if (arg.startswith("--sequence=")) or (arg.startswith("--sequences=")) or (arg.startswith("--ref=")):
			referenceFilename = argVal
			if  ((referenceFilename.endswith(".2bit")) \
			  or (referenceFilename.endswith(".fa.gz")) \
			  or (referenceFilename.endswith(".fasta.gz")) \
			  or (referenceFilename.endswith(".fsa_nt.gz")) \
			  or (referenceFilename.endswith(".fa")) \
			  or (referenceFilename.endswith(".fasta")) \
			  or (referenceFilename.endswith(".fsa_nt"))):
				pass
			else:
				usage("unrecognized file extension in %s" % arg)
			referenceFilenames += [referenceFilename]
		elif (arg.startswith("--alias:")):
			argVal = arg.split(":",1)[1]
			for argField in argVal.split(","):
				(alias,name) = argField.split("=",1)
				assert (alias not in aliasToName)
				aliasToName[alias] = name
		elif (arg.startswith("--head=")):
			headLimit = int_with_unit(argVal)
		elif (arg.startswith("--progress=")):
			reportProgress = int_with_unit(argVal)
		elif (arg in ["--version","--v","--V","-version","-v","-V"]):
			exit("%s, version %s" % (programName,programVersion))
		elif (arg == "--debug"):
			debug += ["debug"]
		elif (arg.startswith("--debug=")):
			debug += argVal.split(",")
		elif (arg in ["--refdebug","--2bitdebug"]):
			refDebug += ["debug"]
		elif (arg.startswith("--refdebug=")) or (arg.startswith("--2bitdebug=")):
			refDebug += argVal.split(",")
		elif (arg.startswith("--")):
			usage("unrecognized option: %s" % arg)
		else:
			usage("unrecognized option: %s" % arg)

	if (referenceFilenames == []):
		usage("you need to provide a reference genome file")

	if (len(referenceFilenames) > 1):
		usage("sorry, support for more than one reference genome file had not been implemented yet")

	# open the reference lookup (if we have one); this will map genomic
	# positions or segments to the corresponding nucleotide(s)

	refLookup = None

	if (referenceFilenames != []):
		referenceFilename = referenceFilenames[0]

		if (referenceFilename.endswith(".2bit")):
			if (TwoBitFile == None):
				assert (False), \
					   "two bit format support is not installed (needed for \"%s\")" % referenceFilename
			refLookup = TwoBitFile(referenceFilename,unmask=False,
		                           debug=refDebug)
		elif (referenceFilename.endswith(".gz")) or (referenceFilename.endswith(".gzip")):
			# (assume it is gzipped fasta)
			if (FastaFile == None):
				assert (False), \
					   "fasta format support is not installed (needed for \"%s\")" % referenceFilename
			refLookup = FastaFile(referenceFilename,unmask=False,
			                      ignoreFai=True,
			                      debug=refDebug)
		else:
			# (assume it is fasta)
			if (FastaFile == None):
				assert (False), \
					   "fasta format support is not installed (needed for \"%s\")" % referenceFilename
			refLookup = FastaFile(referenceFilename,unmask=False,
			                      ignoreFai=("ignorefai" in debug),
			                      debug=refDebug)

	# open the alignment table, as an iterator
	#
	# nota bene: we require the cigarx field but allow "cigar" as an alias for
	#            it; in this way we'll guarantee we have one or the other; and
	#            either is adequate for our purposes; note that this will be
	#            a.cigarx in an alignment object regardless of which the user
	#            provided

	requiredColumns    = ("name1","zstart1","end1",
	                      "name2","zstart2+","end2+",
	                      "strand",
	                      "cigarx")
	nonRequiredColumns = ("score","text1","text2")
	columnAliases      = {"strand2" : "strand",
	                      "s"       : "strand",
	                      "s2"      : "strand",
	                      "cigar"   : "cigarx"}
	for alias in aliasToName:
		columnAliases[alias] = aliasToName[alias]

	t = AlignmentTable.from_file(stdin,
	                             requiredColumns=requiredColumns,
	                             nonRequiredColumns=nonRequiredColumns,
	                             columnAliases=columnAliases)

	# process the alignments

	headerHasBeenWritten = False

	alignmentNumber = 0
	for a in t:
		alignmentNumber += 1

		if (headLimit != None) and (alignmentNumber > headLimit):
			print("limit of %s alignments reached" % ("{:,}".format(headLimit)),file=stderr)
			break
		if (reportProgress != None):
			if (alignmentNumber == 1) or (alignmentNumber % reportProgress == 0):
				print("processing alignment %s" % "{:,}".format(alignmentNumber),file=stderr)

		if (not headerHasBeenWritten):
			header = ["##maf","version=1"]
			header += ["scoring=lastz_defaults"]
			print(" ".join(header))
			print("# this file generated by %s version %s, converting from lastz tabular format" \
			     % (programName,programVersion))
			print("#")
			print("# gap_open_penalty   = 400")
			print("# gap_extend_penalty = 30")
			print("#        A    C    G    T")
			print("#   A   91 -114  -31 -123")
			print("#   C -114  100 -125  -31")
			print("#   G  -31 -125  100 -114")
			print("#   T -123  -31 -114   91")
			headerHasBeenWritten = True

		if ("linenumber" in debug):
			print("# line %d" % a.lineNumber)
		alignment_to_maf(a,refLookup)


# alignment_to_maf--
#	print a (pairwise) alignment in maf format

def alignment_to_maf(a,refLookup):
	seq1 = refLookup[a.name1]
	assert (seq1 != None), "no sequence has been provided for %s" % (a.name1)
	seq2 = refLookup[a.name2]
	assert (seq2 != None), "no sequence has been provided for %s" % (a.name2)
	(srcSize1,srcSize2) = (len(seq1),len(seq2))

	if (hasattr(a,"text1")) and (hasattr(a,"text2")):
		(text1,text2) = (a.text1,a.text2)
	else:
		cigarOps = split_cigar(a.cigarx)
		(text1,text2) = construct_alignment_text(a,cigarOps,refLookup,refLookup)

	if (hasattr(a,"score")):
		score = a.score
	else:
		score = alignment_score(text1,text2)

	start1 = a.start1
	size1  = a.end1 - a.start1

	start2 = a.start2 if (a.strand == "+") else (srcSize2 - a.end2)
	size2  = a.end2 - a.start2

	srcW     = max(len(a.name1)      ,len(a.name2))
	startW   = max(len(str(start1  )),len(str(start2  )))
	sizeW    = max(len(str(size1   )),len(str(size2   )))
	srcSizeW = max(len(str(srcSize1)),len(str(srcSize2)))

	print("a score=%s" % score)
	print("s %-*s %*s %*s %s %*s %s" \
	    % (srcW,a.name1,
	       startW,start1,
	       sizeW,size1,
	       "+",
	       srcSizeW,srcSize1,
	       text1))
	print("s %-*s %*s %*s %s %*s %s" \
	    % (srcW,a.name2,
	       startW,start2,
	       sizeW,size2,
	       a.strand,
	       srcSizeW,srcSize2,
	       text2))
	print("")


# int_with_unit--
#	Parse a string as an integer, allowing unit suffixes

def int_with_unit(s):
	if (s.endswith("K")):
		multiplier = 1000
		s = s[:-1]
	elif (s.endswith("M")):
		multiplier = 1000 * 1000
		s = s[:-1]
	elif (s.endswith("G")):
		multiplier = 1000 * 1000 * 1000
		s = s[:-1]
	else:
		multiplier = 1

	try:               return          int(s)   * multiplier
	except ValueError: return int(ceil(float(s) * multiplier))


if __name__ == "__main__": main()
